/*
 * ----------------------------------------------------------------
 * ircproxy - Connection Logging Functions
 * ----------------------------------------------------------------
 * Copyright (C) 1997-2009 Jonas Kvinge
 *
 * This file is part of ircproxy.
 *
 * ircproxy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ircproxy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ircproxy.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Additional permission under GNU GPL version 3 section 7
 *
 * If you modify ircproxy, or any covered work, by linking or
 * combining it with openssl (or a modified version of that library),
 * containing parts covered by the terms of the OpenSSL License and the
 * SSLeay License, the licensors of ircproxy grant you additional
 * permission to convey the resulting work.
 *
 * $Id: conn_log.c 54 2009-03-18 18:23:29Z jonasio $
 *
 */

#define CONN_LOG_C

#define NEED_SYS_TYPES_H 1		/* Extra types */
#define NEED_SYS_PARAM_H 1		/* Some systems need this */
#define NEED_LIMITS_H 0			/* Kernel limits */
#define NEED_STDARG_H 1			/* va_list, etc */
#define NEED_ERRNO_H 1			/* errno */
#define NEED_CTYPE_H 1			/* isdigit(), etc */
#define NEED_NETINET_IN_H 1		/* in_addr, sockaddr_in, etc */
#define NEED_ARPA_INET_H 0		/* inet_ntoa(), inet_aton(), etc */
#define NEED_STDIO_H 1			/* Standard C UNIX functions */
#define NEED_STDLIB_H 1			/* malloc(), exit(), atoi(), etc */
#define NEED_TIME_H 1			/* time(), etc */
#define NEED_SYSCTL_H 0			/* sysctl(), etc */
#define NEED_SYS_STAT_H 1		/* chmod(), mkdir(), etc */
#define NEED_SYS_UIO_H 0		/* iovec, etc */
#define NEED_FCNTL_H 1			/* open(), creat(), fcntl(), etc */
#define NEED_SYS_IOCTL_H 0		/* ioctl(), etc */
#define NEED_SYS_FILIO_H 0		/* Solaris need this for ioctl(), etc */
#define NEED_UNISTD_H 1			/* Unix standard functions */
#define NEED_STRING_H 1			/* C string functions */
#define NEED_SIGNAL_H 0			/* Signal functions */
#define NEED_SYS_SOCKET_H 0		/* Socket functions */
#define NEED_NETDB_H 0			/* Network database functions */
#define NEED_ARPA_NAMESER_H 0		/* Nameserver definitions */
#define NEED_GETUSERPW_HEADERS 0 	/* Functions to retrive system passwords */
#define NEED_ARES 0			/* Functions needed for ares */
#define NEED_SSL 1			/* Needed for SSL support */

#include "includes.h"
#include "irc.h"

#include "conf.h"

#include "conn.h"
#include "conn_log.h"

#include "chan.h"

extern struct Conf_Struct ConfS;
extern unsigned short int Root;

/* CONN_ADDLOG - JONAS (20.01.2008) */

struct ConnLog_Struct *conn_addlog(struct Conn_Struct *ConnS, unsigned short int LogType, const char *const NamePT, const char *const LogFilePT, const char *const LogFileFullPathPT) {

  struct ConnLog_Struct *ConnLog = NULL;
  struct ConnLog_Struct *ConnLog_NEW = NULL;

  assert(ConnS != NULL);
  assert(NamePT != NULL);
  assert(LogFilePT != NULL);
  assert(LogFileFullPathPT != NULL);

  ConnLog = conn_getlog(ConnS, LogFilePT);
  if (ConnLog != NULL) {
    aerrno = AEEXISTS;
    return(ConnLog);
  }

  ConnLog_NEW = malloc(sizeof(struct ConnLog_Struct));
  if (ConnLog_NEW == NULL) {
    aerrno = AEMALLOC;
    return(NULL);
  }
  memset(ConnLog_NEW, 0, sizeof(struct ConnLog_Struct));

  ConnLog_NEW->LogType = LogType;
  ConnLog_NEW->Name = strdup(NamePT);
  ConnLog_NEW->LogFile = strdup(LogFilePT);
  ConnLog_NEW->LogFileFullPath = strdup(LogFileFullPathPT);

  if ((ConnLog_NEW->Name == NULL) || (ConnLog_NEW->LogFile == NULL) || (ConnLog_NEW->LogFileFullPath == NULL)) {
    free(ConnLog_NEW->Name);
    free(ConnLog_NEW->LogFile);
    free(ConnLog_NEW->LogFileFullPath);
    free(ConnLog_NEW);
    aerrno = AEMALLOC;
    return(NULL);
  }

  if (ConnS->Log_Head == NULL) {
    ConnS->Log_Head = ConnLog_NEW;
    ConnS->Log_Tail = ConnLog_NEW;
  }
  else {
    ConnLog = ConnS->Log_Tail;
    ConnLog->Next = ConnLog_NEW;
    ConnLog_NEW->Prev = ConnLog;
    ConnS->Log_Tail = ConnLog_NEW;
  }

  ConnS->NumLogs++;

  aerrno = AESUCCESS;
  return(ConnLog_NEW);

}

/* CONN_REMLOG FUNCTION - JONAS (20.01.2008) */

void conn_remlog(struct Conn_Struct *ConnS, struct ConnLog_Struct *ConnLog) {

  assert(ConnS != NULL);
  assert(ConnLog != NULL);

  if (ConnLog->Prev == NULL) { ConnS->Log_Head = ConnLog->Next; }
  else { ConnLog->Prev->Next = ConnLog->Next; }

  if (ConnLog->Next == NULL) { ConnS->Log_Tail = ConnLog->Prev; }
  else { ConnLog->Next->Prev = ConnLog->Prev; }

  free(ConnLog->Name);
  free(ConnLog->LogFile);
  free(ConnLog->LogFileFullPath);
  free(ConnLog);

  ConnS->NumLogs--;

}

/* CONN_GETLOG FUNCTION - JONAS (20.01.2008) */

struct ConnLog_Struct *conn_getlog(struct Conn_Struct *ConnS, const char *const LogFilePT) {

  struct ConnLog_Struct *ConnLog = NULL;

  assert(ConnS != NULL);
  assert(LogFilePT != NULL);

  for (ConnLog = ConnS->Log_Head ; ConnLog != NULL ; ConnLog = ConnLog->Next) {
    if (strcasecmp(ConnLog->LogFile, LogFilePT) == FALSE) {
      aerrno = AESUCCESS;
      return(ConnLog);
    }
  }
  aerrno = AENOMATCH;
  return(NULL);

}

/* CONN_REMLOGS FUNCTION - JONAS (20.01.2008) */

void conn_destroylogs(struct Conn_Struct *ConnS) {

  while (ConnS->Log_Head != NULL) { conn_remlog(ConnS, ConnS->Log_Head); }

}

/* CONN_LOG FUNCTION - JONAS (18.07.2001) */

unsigned short int conn_log(struct Conn_Struct *ConnS, struct Chan_Struct *ChanS, const char *const ChanPT, const char *const NickPT, const char *const UserPT, const char *const HostPT, const char *const NUHPT, const char *const CommandPT, const char *const ParamPT, const char *const MessagePT) {

  signed long int Result = 0;
  char *HomeDirPT = NULL;
  char LogFileName[FILELEN+1] = "";
  char LogPath[FILELEN+1] = "";
  char LogFileNamePath[FILELEN+1] = "";
  char **FilePTT = NULL;
  FILE *FileFP = NULL;
  FILE **FileFPP = &FileFP;
#if 0
  struct tm *TMPT = NULL;
  char Date[LINELEN+1] = "";
#endif
  unsigned short int LogErrors = 0;
  char *LogFileNamePT = NULL;

  assert(ConnS != NULL);

  Conn_ClearLogHomeDir(ConnS);

  if ((ConfS.UnixPasswd == TRUE) && (Root == TRUE)) {
    HomeDirPT = sysgethomedirfromuser(ConnS->User);
    if (HomeDirPT == NULL) {
      ConnLogError("Unable to log messages for connection %s: Unable to retrive home directory for user \"%s\"", ConnS->Name, ConnS->User);
      return(ERROR);
    }
    if (sysseteuidbyuser(ConnS->User) != SUCCESS) {
      ConnLogError("Unable to log messages for connection %s: Unable to set effective user ID to \"%s\"", ConnS->Name, ConnS->User);
      return(ERROR);
    }
    if (access(HomeDirPT, X_OK) != SUCCESS) {
      ConnLogError("Unable to log messages for connection %s: Unable to access home directory for user \"%s\"", ConnS->Name, ConnS->User);
      sysseteuidnormal();
      return(ERROR);
    }
    Conn_SetLogHomeDir(ConnS);
    snprintf(LogPath, FILELEN+1, "%s/ircproxy-logs", HomeDirPT);
  }
  else { snprintf(LogPath, FILELEN+1, "%s", ConfS.LogPath); }

  if (access(LogPath, X_OK) == ERROR) {
    Result = mkdir(LogPath, (S_IRWXU|S_IXGRP|S_IXOTH));
    if ((Result <= ERROR) && (errno != EEXIST)) {
      sysprint(BITMASK_ERROR, "Unable to create directory %s: [%d] %s", LogPath, errno, strerror(errno));
      if (Conn_IsLogHomeDir(ConnS)) { sysseteuidnormal(); }
      return(ERROR);
    }
  }

#if 0
  TMPT = localtime(&NOW);
  strftime(Date, LINELEN, "%Y%m%d", TMPT);
#endif

  VariableDataS.ConnS = ConnS;
  VariableDataS.ChanS = ChanS;
  VariableDataS.Chan = ChanPT;
  VariableDataS.Nick = NickPT;
  VariableDataS.User = UserPT;
  VariableDataS.Host = HostPT;
  VariableDataS.NUH = NUHPT;
  VariableDataS.Command = CommandPT;
  VariableDataS.Param = ParamPT;
  VariableDataS.Message = MessagePT;

  if (CommandPT == NULL) {
    LogFileNamePT = conn_log_variable(ConfS.RawLogFileFormat, NULL);
#if 0
    snprintf(LogFileName, FILELEN+1, "%s-%s-%s.log", ConnS->Name, ConnS->ServerHostName, Date);
#endif
    FilePTT = &ConnS->RawLogFile;
    FileFPP = &ConnS->RawLogFileFP;
  }
  else {
    if (ChanPT == NULL) {
      LogFileNamePT = conn_log_variable(ConfS.PrivLogFileFormat, NULL);
#if 0
      snprintf(LogFileName, FILELEN+1, "%s-%s.log", ConnS->Name, Date);
#endif
      FilePTT = &ConnS->PrivLogFile;
      FileFPP = &ConnS->PrivLogFileFP;
    }
    else {
      LogFileNamePT = conn_log_variable(ConfS.ChanLogFileFormat, NULL);
#if 0
      snprintf(LogFileName, FILELEN+1, "%s-%s-%s.log", ConnS->Name, ChanPT, Date);
#endif
      if (ChanS == NULL) {
        FileFPP = &FileFP;
      }
      else {
        FilePTT = &ChanS->LogFile;
        FileFPP = &ChanS->LogFileFP;
      }
    }
  }
  strcpy(LogFileName, LogFileNamePT);
  FREE(LogFileNamePT);
  strfilefix(LogFileName);
  snprintf(LogFileNamePath, FILELEN+1, "%s/%s", LogPath, LogFileName);

  if ((*FilePTT != NULL) && (strcmp(*FilePTT, LogFileName) != FALSE)) {
    FREE(*FilePTT);
    if (*FileFPP != NULL) {
      fclose(*FileFPP);
      *FileFPP = NULL;
    }
  }

  do {

    if (*FileFPP == NULL) {
      *FileFPP = fopen(LogFileNamePath, "a");
      if (*FileFPP == NULL) {
        ConnLogError("Unable to open %s for writing: [%d] %s", LogFileNamePath, errno, strerror(errno));
        if (Conn_IsLogHomeDir(ConnS)) { sysseteuidnormal(); }
        return(ERROR);
      }
      if (FilePTT != NULL) *FilePTT = strrealloc(*FilePTT, LogFileName);
      Result = chmod(LogFileNamePath, S_IRUSR|S_IWUSR);
      if (Result <= ERROR) {
        ConnLogError("Unable to set file permission for %s: [%d] %s", LogFileNamePath, errno, strerror(errno));
#if 0
        fclose(*FileFPP);
        if (FilePTT != NULL) FREE(*FilePTT);
        if (Conn_IsLogHomeDir(ConnS)) { sysseteuidnormal(); }
        return(ERROR);
#endif
      }
      Result = fprintf(*FileFPP, "%s *** Logging started.%s", dtstamp(), LINEFEED);
    }

    if (CommandPT == NULL) { Result = fprintf(*FileFPP, "%s *** %s%s", dtstamp(), MessagePT, LINEFEED); }
    else {
      assert(CommandPT != NULL);
      assert(NUHPT != NULL);
      assert(NickPT != NULL);
      assert(UserPT != NULL);
      assert(HostPT != NULL);
      if (strcasecmp(CommandPT, "PRIVMSG") == FALSE) { Result = fprintf(*FileFPP, "%s <%s> %s%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
      else if (strcasecmp(CommandPT, "NOTICE") == FALSE) { Result = fprintf(*FileFPP, "%s -%s- %s%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
      else if (strcasecmp(CommandPT, "ACTION") == FALSE) { Result = fprintf(*FileFPP, "%s * %s %s%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
      else if (strcasecmp(CommandPT, "CTCP") == FALSE) {
        if (MessagePT == NULL)  { Result = fprintf(*FileFPP, "%s %s CTCP %s%s", dtstamp(), NUHPT, ParamPT, LINEFEED); }
        else { Result = fprintf(*FileFPP, "%s %s CTCP %s %s%s", dtstamp(), NUHPT, ParamPT, MessagePT, LINEFEED); }
      }
      else if (ChanPT != NULL) {
        if (strcasecmp(CommandPT, "NICK") == FALSE) { Result = fprintf(*FileFPP, "%s %s (%s@%s) is now known as %s%s", dtstamp(), NickPT, UserPT, HostPT, ParamPT, LINEFEED); }
        else if (strcasecmp(CommandPT, "QUIT") == FALSE) {
          if (MessagePT == NULL) { Result = fprintf(*FileFPP, "%s %s quits%s", dtstamp(), NUHPT, LINEFEED); }
          else { Result = fprintf(*FileFPP, "%s %s quits: %s%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
        }
        else if (strcasecmp(CommandPT, "MODE") == FALSE) { Result = fprintf(*FileFPP, "%s %s sets mode: %s%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }
        else if (strcasecmp(CommandPT, "JOIN") == FALSE) { Result = fprintf(*FileFPP, "%s %s has joined %s%s", dtstamp(), NUHPT, ChanPT, LINEFEED); }
        else if (strcasecmp(CommandPT, "PART") == FALSE) {
          if (MessagePT == NULL) { Result = fprintf(*FileFPP, "%s %s has left %s%s", dtstamp(), NUHPT, ChanPT, LINEFEED); }
          else { Result = fprintf(*FileFPP, "%s %s has left %s: %s%s", dtstamp(), NUHPT, ChanPT, MessagePT, LINEFEED); }
        }
        else if (strcasecmp(CommandPT, "KICK") == FALSE) { Result = fprintf(*FileFPP, "%s %s was kicked by %s: %s%s", dtstamp(), ParamPT, NUHPT, MessagePT, LINEFEED); }
        else if (strcasecmp(CommandPT, "TOPIC") == FALSE) { Result = fprintf(*FileFPP, "%s %s changed topic to: '%s'%s", dtstamp(), NUHPT, MessagePT, LINEFEED); }

      }
    }

    if (Result <= ERROR) {
      ++LogErrors;
      fclose(*FileFPP);
      *FileFPP = NULL;
      if (FilePTT != NULL) FREE(*FilePTT);
    }
    else { break; }

  }
  while (LogErrors < 2);

  if (FileFPP != NULL) fflush(*FileFPP);
  if (FileFPP == &FileFP) fclose(*FileFPP);

  if (Conn_IsLogHomeDir(ConnS)) { sysseteuidnormal(); }

  if (LogErrors < 2) {
    if (CommandPT == NULL) { conn_addlog(ConnS, CONN_LOG_TYPE_RAW, ConnS->ServerHostName, LogFileName, LogFileNamePath); }
    else {
      if (ChanPT == NULL) { conn_addlog(ConnS, CONN_LOG_TYPE_PRIV, ConnS->Name, LogFileName, LogFileNamePath); }
      else { conn_addlog(ConnS, CONN_LOG_TYPE_CHAN, ChanPT, LogFileName, LogFileNamePath); }
    }
  }

#if 0
  for (Index = 0 ; Index < ConnS->Logs ; ++Index) {
    if (strcasecmp(ConnS->LogFile, LogFileName) == FALSE) { break; }
  }
  if (Index == ConnS->Logs) {
    ++ConnS->Logs;
    TempPT = realloc(ConnS->LogFile, (sizeof(char *) * ConnS->Logs));
    if (TempPT == NULL) {
      sysprint(BITMASK_ERROR, "Memory allocation failure: [%d] %s", errno, strerror(errno));
      return(ERROR);
    }
    ConnS->LogFile = TempPT;
    ConnS->LogFile[Index] = strdup(LogFileNamePath);
  }
#endif

  return(SUCCESS);

}

/* CONN_RAWLOG FUNCTION - JONAS (01.07.2000) */

void conn_rawlog(struct Conn_Struct *ConnS, const char *const LinePT, ...) {

  char Line[IRC_MSGLEN+1] = "";
  va_list Args;

  assert(ConnS != NULL);
  assert(LinePT != NULL);

  va_start(Args, LinePT);
  vsnprintf(Line, IRC_MSGLEN+1, LinePT, Args);
  va_end(Args);

  conn_log(ConnS, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, Line);

}

/* CONN_LOG_VARIABLE FUNCTION - JONAS (30.09.2001) */

char *conn_log_variable(const char *OldMessagePT, char *NewMessagePT) {

  unsigned long int Len = 0;
  unsigned long int Index = 0;
  char *MessagePT = NULL;
  char *TempPT = NULL;
  const char *VarNamePT = NULL;
  const char *VarDataPT = NULL;

  unsigned long int VSLen = 0;
  unsigned long int VSIndex = 0;

  struct tm *TMPT = NULL;
  char TimeYear[LINELEN+1] = "";
  char TimeMonth[LINELEN+1] = "";
  char TimeWeek[LINELEN+1] = "";
  char TimeDay[LINELEN+1] = "";
  char TimeHour[LINELEN+1] = "";
  char TimeMin[LINELEN+1] = "";
  char TimeSec[LINELEN+1] = "";

  struct Variable_Struct Variable[] = {

    {
      "%mhost",
      HOSTNAME
    },
    {
      "%conn",
      ((VariableDataS.ConnS == NULL) ? NULL : VariableDataS.ConnS->Name)
    },
    {
      "%serverhost",
      ((VariableDataS.ConnS == NULL) ? NULL : VariableDataS.ConnS->ServerHostName)
    },
    {
      "%servername",
      ((VariableDataS.ConnS == NULL) ? NULL : VariableDataS.ConnS->ServerName)
    },
    {
      "%cnick",
      ((VariableDataS.ConnS == NULL) ? NULL : VariableDataS.ConnS->Nick)
    },
    {
      "%cuser",
      ((VariableDataS.ConnS == NULL) ? NULL : VariableDataS.ConnS->User)
    },
    {
      "%chost",
      ((VariableDataS.ConnS == NULL) ? NULL : VariableDataS.ConnS->Host)
    },
    {
      "%anick",
      ((VariableDataS.ConnS == NULL) ? NULL : VariableDataS.ConnS->IRCNick)
    },
    {
      "%auser",
      ((VariableDataS.ConnS == NULL) ? NULL : VariableDataS.ConnS->User)
    },
    {
      "%ahost",
      ((VariableDataS.ConnS == NULL) ? NULL : VariableDataS.ConnS->Host)
    },
    {
      "%nick",
      VariableDataS.Nick
    },
    {
      "%user",
      VariableDataS.User
    },
    {
      "%host",
      VariableDataS.Host
    },
    {
      "%nuh",
      VariableDataS.NUH
    },
    {
      "%chan",
      VariableDataS.Chan
    },
    {
      "%y",
      TimeYear
    },
    {
      "%m",
      TimeMonth
    },
    {
      "%w",
      TimeWeek
    },
    {
      "%d",
      TimeDay
    },
    {
      "%h",
      TimeHour
    },
    {
      "%n",
      TimeMin
    },
    {
      "%s",
      TimeSec
    },

  };

  VSLen = (sizeof(Variable) / sizeof(*Variable));

  TMPT = localtime(&NOW);
  strftime(TimeYear, LINELEN, "%Y", TMPT);
  strftime(TimeMonth, LINELEN, "%m", TMPT);
  strftime(TimeWeek, LINELEN, "%W", TMPT);
  strftime(TimeDay, LINELEN, "%d", TMPT);
  strftime(TimeHour, LINELEN, "%H", TMPT);
  strftime(TimeMin, LINELEN, "%M", TMPT);
  strftime(TimeSec, LINELEN, "%S", TMPT);

  for (Len = 1 ; *OldMessagePT != 0 ;) {

    TempPT = strchr(OldMessagePT, '%');
    if (TempPT == NULL) {
      Len += strlen(OldMessagePT);
      MessagePT = realloc(NewMessagePT, Len);
      if (MessagePT == NULL) {
        free(NewMessagePT);
        return(NULL);
      }
      NewMessagePT = MessagePT;
      MessagePT += Index;
      strcpy(MessagePT, OldMessagePT);
      OldMessagePT += strlen(OldMessagePT);
      Index += strlen(OldMessagePT);
    }
    else {

      if (TempPT != OldMessagePT) {

        Len += (TempPT - OldMessagePT);
        MessagePT = realloc(NewMessagePT, Len);
        if (MessagePT == NULL) {
          free(NewMessagePT);
          return(NULL);
        }
        NewMessagePT = MessagePT;
        MessagePT += Index;
        strncpy(MessagePT, OldMessagePT, (TempPT - OldMessagePT));
        Index += (TempPT - OldMessagePT);
        OldMessagePT = TempPT;

      }
      VarNamePT = NULL;
      VarDataPT = NULL;
      for (VSIndex = 0 ; VSIndex < VSLen ; VSIndex++) {
        if ((Variable[VSIndex].Data != NULL) && (strncasecmp(OldMessagePT, Variable[VSIndex].Name, strlen(Variable[VSIndex].Name)) == FALSE)) {
          VarNamePT = Variable[VSIndex].Name;
          VarDataPT = Variable[VSIndex].Data;
          break;
        }
      }
      if ((VarNamePT != NULL) && (VarDataPT != NULL)) {
        Len += strlen(VarDataPT);
        MessagePT = realloc(NewMessagePT, Len);
        if (MessagePT == NULL) {
          free(NewMessagePT);
          return(NULL);
        }
        NewMessagePT = MessagePT;
        MessagePT += Index;
        strcpy(MessagePT, VarDataPT);
        OldMessagePT += strlen(VarNamePT);
        Index += strlen(VarDataPT);
      }
      else {
        Len += strlen(OldMessagePT);
        MessagePT = realloc(NewMessagePT, Len);
        if (MessagePT == NULL) {
          free(NewMessagePT);
          return(NULL);
        }
        NewMessagePT = MessagePT;
        MessagePT += Index;
        strcpy(MessagePT, OldMessagePT);
        OldMessagePT += strlen(OldMessagePT);
        Index += strlen(OldMessagePT);
      }
    }
  }

  VariableDataS.ConnS = NULL;
  VariableDataS.ChanS = NULL;
  VariableDataS.Chan = NULL;
  VariableDataS.Nick = NULL;
  VariableDataS.User = NULL;
  VariableDataS.Host = NULL;
  VariableDataS.NUH = NULL;
  VariableDataS.Command = NULL;
  VariableDataS.Param = NULL;
  VariableDataS.Message = NULL;

  return(NewMessagePT);

}
