/*
 * ----------------------------------------------------------------
 * libmemcalls - Memory Allocation List Functions
 * ----------------------------------------------------------------
 * Copyright (C) 1997-2009 Jonas Kvinge
 *
 * This file is part of libmemcalls.
 *
 * libmemcalls is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * libmemcalls is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * $Id: memlist.c 54 2009-03-18 18:23:29Z jonasio $
 *
 */

#define MEMLIST_C

#define NEED_SYS_TYPES_H 1		/* Extra types */
#define NEED_SYS_PARAM_H 1		/* Some systems need this */
#define NEED_LIMITS_H 0			/* Kernel limits */
#define NEED_STDARG_H 1			/* va_list, etc */
#define NEED_ERRNO_H 1			/* errno */
#define NEED_CTYPE_H 1			/* isdigit(), etc */
#define NEED_NETINET_IN_H 1		/* in_addr, sockaddr_in, etc */
#define NEED_ARPA_INET_H 0		/* inet_ntoa(), inet_aton(), etc */
#define NEED_STDIO_H 1			/* Standard C UNIX functions */
#define NEED_STDLIB_H 1			/* malloc(), exit(), atoi(), etc */
#define NEED_TIME_H 1			/* time(), etc */
#define NEED_SYSCTL_H 0			/* sysctl(), etc */
#define NEED_SYS_STAT_H 0		/* chmod(), mkdir(), etc */
#define NEED_SYS_UIO_H 0		/* iovec, etc */
#define NEED_FCNTL_H 1			/* open(), creat(), fcntl(), etc */
#define NEED_SYS_IOCTL_H 0		/* ioctl(), etc */
#define NEED_SYS_FILIO_H 0		/* Solaris need this for ioctl(), etc */
#define NEED_UNISTD_H 1			/* Unix standard functions */
#define NEED_STRING_H 1			/* C string functions */
#define NEED_SIGNAL_H 0			/* Signal functions */
#define NEED_SYS_SOCKET_H 0		/* Socket functions */
#define NEED_NETDB_H 0			/* Network database functions */
#define NEED_ARPA_NAMESER_H 0		/* Nameserver definitions */
#define NEED_GETUSERPW_HEADERS 0 	/* Functions to retrive system passwords */
#define NEED_ARES 0			/* Functions needed for ares */
#define NEED_SSL 1			/* Needed for SSL support */

#include "includes.h"

#include "memcalls_private.h"
#include "memlist_private.h"

/* VARIABLES - JONAS (21.09.2001) */

extern struct MemTable_Struct *MemTable_Head;
extern struct MemTable_Struct *MemTable_Tail;
extern unsigned long int G_Allocations;
extern size_t MemPrefixSize;
extern size_t MemPostfixSize;

extern char *TMPSTRPT;

extern struct PrintFile_Struct PrintFileS[];
extern unsigned short int PrintFileS_Len;

/* MEM_LIST - JONAS (21.09.2001) */

void mem_list(void) {

  signed long int Result = 0;
  struct MemTable_Struct *MemTableS = NULL;
  unsigned long int Allocations = 0;
  unsigned long int AllocationsSize = 0;
  char PTRZSTR[MEMPTRZSTRLEN+1] = "";
  char BYTESTR[MEMBYTESTRLEN+1] = "";
  char FILESTR[MEMFILESTRLEN+1] = "";
  char LINESTR[MEMLINESTRLEN+1] = "";
  char TIMESTR[MEMTIMESTRLEN+1] = "";

  for (MemTableS = MemTable_Head ; MemTableS != NULL ; MemTableS = MemTableS->Next) {

    const void *PrefixPTR = MemTableS->PTR - MemPrefixSize;
    const void *PostfixPTR = MemTableS->PTR + MemTableS->Size;

    Allocations++;
    AllocationsSize += MemTableS->Size;

    if (strncmp(PrefixPTR, MALLOC_PREFIX, MALLOC_PREFIX_LEN) != 0) {
      sysprint(BITMASK_ERROR, "Magic memory prefix \"%s\" corrupt on pointer %p allocated in function %s (%s:%d).", MALLOC_PREFIX, MemTableS->PTR, MemTableS->Function, MemTableS->File, MemTableS->Line);
      abort();
    }

    if (strncmp(PostfixPTR, MALLOC_POSTFIX, MALLOC_POSTFIX_LEN) != 0) {
      sysprint(BITMASK_ERROR, "Magic memory postfix \"%s\" corrupt on pointer %p allocated in function %s (%s:%d).", MALLOC_POSTFIX, MemTableS->PTR, MemTableS->Function, MemTableS->File, MemTableS->Line);
      abort();
    }

    memset(&PTRZSTR, 32, MEMPTRZSTRLEN);
    memset(&BYTESTR, 32, MEMBYTESTRLEN);
    memset(&FILESTR, 32, MEMFILESTRLEN);
    memset(&LINESTR, 32, MEMLINESTRLEN);
    memset(&TIMESTR, 32, MEMTIMESTRLEN);

    Result = snprintf(PTRZSTR, MEMPTRZSTRLEN+1, "%p", MemTableS->PTR);
    if (Result > 0) { PTRZSTR[Result] = 32; }
    PTRZSTR[MEMPTRZSTRLEN] = 0;

    Result = snprintf(BYTESTR, MEMBYTESTRLEN+1, "%ld", (PRINT_SIZE_T) MemTableS->Size);
    if (Result > 0) { BYTESTR[Result] = 32; }
    BYTESTR[MEMBYTESTRLEN] = 0;

    Result = snprintf(FILESTR, MEMFILESTRLEN+1, "%s", MemTableS->File);
    if (Result > 0) { FILESTR[Result] = 32; }
    FILESTR[MEMFILESTRLEN] = 0;

    Result = snprintf(LINESTR, MEMLINESTRLEN+1, "%d", MemTableS->Line);
    if (Result > 0) { LINESTR[Result] = 32; }
    LINESTR[MEMLINESTRLEN] = 0;

    Result = snprintf(TIMESTR, MEMTIMESTRLEN+1, "%ld", (PRINT_TIME_T) MemTableS->Time);
    if (Result > 0) { TIMESTR[Result] = 32; }
    TIMESTR[MEMTIMESTRLEN] = 0;

    sysprint(BITMASK_ERROR, "Memory allocation: Hex: %s Size: %s File: %s Line: %s Time: %s", PTRZSTR, BYTESTR, FILESTR, LINESTR, TIMESTR);

  }

  sysprint(BITMASK_MAIN, "Memory allocations: %ld size: %ld", Allocations, AllocationsSize);

}



